// Copyright (c) 2011 Public Software Group e. V., Berlin, Germany
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to
// deal in the Software without restriction, including without limitation the
// rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
// sell copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.

/*

hdrcompose is a simple command line utility to create HDR images.

Version: 0.1

Dependencies:
  ImageMagick  available at http://www.imagemagick.org/
  libtiff      available at http://www.libtiff.org/

Compile with:
gcc -O3 -ffast-math -march=native -I/usr/local/include -L/usr/local/lib -lm -ltiff -o hdrcompose hdrcompose.c `MagickWand-config --cflags --cppflags --ldflags --libs`

Replace /usr/local/include and /usr/local/lib as neccessary for libtiff.


Usage: hdrcompose <curve> <file1> <ev1> [<file2> <ev2> ...] <output_tiff_file>

Supported curves: "linear", "srgb", "adobe"
If you are unsure about which response curve to use, then try "srgb".

Example:
hdrcompose srgb short-exposure.jpg -2 medium.jpg 0 long-exposure.jpg 2 hdr.tiff

*/


#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include <wand/MagickWand.h>
#include <tiffio.h>

#define limit(x) (((x) > 1.0) ? 1.0 : (x))
#define weight_func(x) (1.0 - cos(limit(3.0*M_PI * (x))))

double remove_gamma_linear(double v) {
  return v;
}

double remove_gamma_srgb(double v) {
  if (v <= 0.03928) return v / 12.92;
  else return pow((v + 0.055) / 1.055, 2.4);
}

double remove_gamma_adobe(double v) {
  return pow(v, 2.19921875);
}

typedef double (*remove_gamma_fptr)(double);

int main(int argc, char **argv) {
  int input_file_count = 0;
  int width, height;
  MagickWand **input_images;
  MagickBooleanType magick_status;
  float **input_row_buffers;
  float *output_image;
  double *ev_values;
  double *ev_factors;
  double lowest_ev;
  double highest_ev;
  int i, j;
  size_t x, y;
  remove_gamma_fptr remove_gamma;
  TIFF* tif;
  if (argc < 5 || (argc-3) % 2) {
    char *command;
    if (argc >= 1) command = argv[0];
    else command = "executable";
    fprintf(stderr, "Usage: %s <curve> <file1> <ev1> [<file2> <ev2> ...] <output_tiff_file>\n", command);
    fprintf(stderr, "Supported curves: \"linear\", \"srgb\", \"adobe\"\n");
    exit(1);
  }
  if      (!strcasecmp(argv[1], "linear")) remove_gamma = remove_gamma_linear;
  else if (!strcasecmp(argv[1], "srgb"  )) remove_gamma = remove_gamma_srgb;
  else if (!strcasecmp(argv[1], "adobe" )) remove_gamma = remove_gamma_adobe;
  else {
    fprintf(stderr, "Curve \"%s\" is not supported.\n", argv[1]);
    fprintf(stderr, "Supported curves: \"linear\", \"srgb\", \"adobe\"\n");
    exit(1);
  }
  input_file_count = (argc-3) / 2;
  input_images = malloc(input_file_count * sizeof(MagickWand *));
  if (!input_images) {
    fprintf(stderr, "Failed to allocate memory.\n");
    exit(1);
  }
  input_row_buffers = malloc(input_file_count * sizeof(float *));
  if (!input_images) {
    fprintf(stderr, "Failed to allocate memory.\n");
    exit(1);
  }
  ev_values = malloc(input_file_count * sizeof(double));
  if (!ev_values) {
    fprintf(stderr, "Failed to allocate memory.\n");
    exit(1);
  }
  MagickWandGenesis();
  for (i=0; i<input_file_count; i++) {
    char *filename;
    char *ev_string;
    double ev_value;
    char *endptr;
    int current_width, current_height;
    filename = argv[2+2*i+0];
    ev_string = argv[2+2*i+1];
    ev_value = strtod(ev_string, &endptr);
    if (*endptr) {
      fprintf(stderr, "Invalid floating point number: \"%s\"\n", ev_string);
      exit(1);
    }
    input_images[i] = NewMagickWand();
    magick_status = MagickReadImage(input_images[i], filename);
    if (magick_status == MagickFalse) {
      char *description;
      ExceptionType severity;
      description = MagickGetException(input_images[i], &severity);
      fprintf(stderr, "Could not read file \"%s\": %s\n", filename, description);
      exit(1);
    }
    current_width  = MagickGetImageWidth(input_images[i]);
    current_height = MagickGetImageHeight(input_images[i]);
    // Avoid integer overflow: 13377^2 * 3 * sizeof(float) < 2^31-1
    if (current_width > 13377 || current_height > 13377) {
      fprintf(stderr, "Image is too big.\n");
    }
    ev_values[i] = ev_value;
    if (i == 0) {
      width  = current_width;
      height = current_height;
      lowest_ev  = ev_value;
      highest_ev = ev_value;
    } else {
      if (current_width != width || current_height != height) {
        fprintf(stderr, "File dimensions are not matching.\n");
        exit(1);
      }
      if (ev_value < lowest_ev)  lowest_ev  = ev_value;
      if (ev_value > highest_ev) highest_ev = ev_value;
    }
    input_row_buffers[i] = malloc(3 * width * sizeof(float));
    if (!input_row_buffers[i]) {
      fprintf(stderr, "Failed to allocate memory.\n");
      exit(1);
    }
  }
  ev_factors = malloc(input_file_count * sizeof(double));
  if (!ev_factors) {
    fprintf(stderr, "Failed to allocate memory.\n");
    exit(1);
  }
  for (i=0; i<input_file_count; i++) {
    ev_factors[i] = pow(2.0, ev_values[i]);
  }
  output_image = malloc(3 * width * height * sizeof(float));
  if (!output_image) abort();
  for (y=0; y<height; y++) {
    for (i=0; i<input_file_count; i++) {
      magick_status = MagickExportImagePixels(
        input_images[i],
        0, y, width, 1,
        "RGB", FloatPixel, input_row_buffers[i]
      );
      if (magick_status == MagickFalse) {
        char *description;
        ExceptionType severity;
        description = MagickGetException(input_images[i], &severity);
        fprintf(stderr, "Error in MagickExportImagePixels call: %s\n", description);
        exit(1);
      }
    }
    for (x=0; x<width; x++) {
      double sum_red    = 0.0;
      double sum_green  = 0.0;
      double sum_blue   = 0.0;
      double sum_weight = 0.0;
      for (i=0; i<input_file_count; i++) {
        double red, green, blue;
        double v, wt, ws, weight;
        red   = remove_gamma(input_row_buffers[i][3*x+0]);
        green = remove_gamma(input_row_buffers[i][3*x+1]);
        blue  = remove_gamma(input_row_buffers[i][3*x+2]);
        v = 1.0 - pow((1.0-red)*(1.0-green)*(1.0-blue), 1.0/3.0);
        wt = weight_func(v);
        ws = 0.0;
        for (j=0; j<input_file_count; j++) {
          ws += weight_func(v * ev_factors[j] / ev_factors[i]);
        }
        if (ws < 0.000000001) {
          if (v < 0.5) {
            wt = 1.0;
            ws = 0.0;
            for (j=0; j<input_file_count; j++) {
              ws += ev_factors[j] * ev_factors[j]
                  / ev_factors[i] / ev_factors[i];
            }
          } else {
            wt = 1.0;
            ws = 1.0;
          }
        }
        weight = wt / ws;
        sum_red    += weight * red   / ev_factors[i];
        sum_green  += weight * green / ev_factors[i];
        sum_blue   += weight * blue  / ev_factors[i];
        sum_weight += weight;
      }
      output_image[3*(width*y+x)+0] = sum_red   / sum_weight;
      output_image[3*(width*y+x)+1] = sum_green / sum_weight;
      output_image[3*(width*y+x)+2] = sum_blue  / sum_weight;
    }
  }
  for (i=0; i<input_file_count; i++) {
    input_images[i] = DestroyMagickWand(input_images[i]);
    free(input_row_buffers[i]);
  }
  free(input_images);
  MagickWandTerminus();
  free(input_row_buffers);
  tif = TIFFOpen(argv[argc-1], "w");
  if (!tif) exit(1);
  if (!TIFFSetField(tif, TIFFTAG_IMAGEWIDTH, (uint32_t)width)      ) exit(1);
  if (!TIFFSetField(tif, TIFFTAG_IMAGELENGTH, (uint32_t)height)    ) exit(1);
  if (!TIFFSetField(tif, TIFFTAG_ROWSPERSTRIP, (uint32_t)height)   ) exit(1);
  if (!TIFFSetField(tif, TIFFTAG_BITSPERSAMPLE,
    (uint16_t)8*(uint16_t)sizeof(float))                           ) exit(1);
  if (!TIFFSetField(tif, TIFFTAG_COMPRESSION, COMPRESSION_DEFLATE) ) exit(1);
  if (!TIFFSetField(tif, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_RGB)     ) exit(1);
  if (!TIFFSetField(tif, TIFFTAG_SAMPLEFORMAT, SAMPLEFORMAT_IEEEFP)) exit(1);
  if (!TIFFSetField(tif, TIFFTAG_SAMPLESPERPIXEL, (uint16_t)3)     ) exit(1);
  if (!TIFFSetField(tif, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG)) exit(1);
  if (TIFFWriteEncodedStrip(
    tif, 0, output_image,
    3 * width * height * sizeof(float)
  ) < 0) exit(1);
  TIFFClose(tif);
  return 0;
}
